/* AutoExpDat_Evaluator.cpp - MyopicTopics.com Addin - part 4
 * Written By: Jason Weiler
 * This code is in the public domain
 */

#include "DebuggerData.h"
#include "AutoExpDat_Evaluator.h"

unsigned __int64 ConvertTargetValueToLocalValue(unsigned __int64 Value
												, unsigned int uiSizeOfValue
												, bool bIsBigEndianTarget);

bool ReadMemoryHelper(DEBUGHELPER * pHelper
						 , unsigned __int64 dlAddrToRead
						 , DWORD dwSizeToRead
						 , VOID* pBuffer
						 , unsigned __int64 dwAlignment
						 , DWORD * pOutSizeRead);


HRESULT HandleMyType_Helper(DEBUGHELPER *pHelper
									  , char *pResult
									  , size_t maxResult
									  , bool bTargetIsBigEndian);


unsigned __int64 Truncate64BitValueTo32BitValue(unsigned __int64 Value);
unsigned int FlipEndian32(unsigned int Data);
unsigned __int64 FlipEndian64(unsigned __int64 Data);

// *****************************************************************
ADDIN_API HRESULT WINAPI HandleMyType_WinXX( DWORD /*dwAddress*/
									  , DEBUGHELPER *pHelper
									  , int nBase
									  , BOOL bUniStrings
									  , char *pResult
									  , size_t maxResult
									  , DWORD /*reserved*/ )
{
	// Win32 and Win64 platforms are little-endian
	bool bTargetIsBigEndian = false;

	return HandleMyType_Helper(pHelper, pResult, maxResult, bTargetIsBigEndian);
}

ADDIN_API HRESULT WINAPI HandleMyType_Xbox360( DWORD /*dwAddress*/
									  , DEBUGHELPER *pHelper
									  , int nBase
									  , BOOL bUniStrings
									  , char *pResult
									  , size_t maxResult
									  , DWORD /*reserved*/ )
{
	// Xbox is big-endian
	bool bTargetIsBigEndian = true;

	return HandleMyType_Helper(pHelper, pResult, maxResult, bTargetIsBigEndian);
}

HRESULT HandleMyType_Helper(DEBUGHELPER *pHelper
									  , char *pResult
									  , size_t maxResult
									  , bool bTargetIsBigEndian)
{
	DebuggerData const & DbgData = GetStaticDebuggerData();

	// Get the target pointer.
	unsigned __int64 dlTargetAddr = pHelper->GetRealAddress(pHelper);

	if (dlTargetAddr == 0)
	{
		// NULL Address
		strcpy_s(pResult, maxResult, "<MyType == NULL>");
		return S_OK;
	}

	if (DbgData.GlobalStringTableData == 0)
	{
		strcpy_s(pResult, maxResult, "<ERROR: Uninitialized String Table>");
		return S_OK;
	}

	// Create a pointer to the StringIndex member of our structure
	unsigned __int64 dlStringIdxPtr = dlTargetAddr + DbgData.MyTypeOffsetToStringIndex;

	// Query that data from the target process
	DWORD dwNumRead = 0;
	unsigned __int64 dlStringIndex;
	if(!ReadMemoryHelper(pHelper, dlStringIdxPtr, DbgData.SizeOfInteger, &dlStringIndex, 0x400, &dwNumRead))
	{
		strcpy_s(pResult, maxResult, "<ERROR: Failure Reading StringIndex>");
		return S_OK;
	}

	// Convert our StringIndex value into a locally-useable form
	dlStringIndex = ConvertTargetValueToLocalValue(dlStringIndex, DbgData.SizeOfInteger, bTargetIsBigEndian);

	// Do some bounds checking on our StringIndex
	if (dlStringIndex >= DbgData.GlobalStringTableMax)
	{
		sprintf_s(pResult, maxResult, "<ERROR: StringIndex is out of range (%d)>", dlStringIndex);
		return S_OK;
	}

	// Calculate the pointer to the string we're interested in
	unsigned __int64 dlStringTableEntry = DbgData.GlobalStringTableData + (dlStringIndex * DbgData.SizeOfPointer);

	// Read the string pointer out of the GlobalStringTable
	unsigned __int64 dlStringPointer=0;
	if(!ReadMemoryHelper(pHelper, dlStringTableEntry, DbgData.SizeOfPointer, &dlStringPointer, 0x400, &dwNumRead))
	{
		strcpy_s(pResult, maxResult, "<ERROR: Failure Reading StringTable Entry Pointer>");
		return S_OK;
	}

	// Convert our StringTable pointer value into a locally-useable form
	dlStringPointer = ConvertTargetValueToLocalValue(dlStringPointer, DbgData.SizeOfPointer, bTargetIsBigEndian);

	// check NULL
	if (dlStringPointer == 0)
	{
		strcpy_s(pResult, maxResult, "<StringEntry == NULL>");
		return S_OK;
	}

	// Read the string out of the target process' memory.
	if ( FAILED(ReadMemoryHelper(pHelper, dlStringPointer, maxResult, (void*)pResult, 0x400, &dwNumRead)) )
	{
		// If we got nothing, then this is a full-fail
		if (dwNumRead == 0)
		{
			strcpy_s(pResult, maxResult, "<ERROR: Failure Reading StringTable Entry String Data>");
			return S_OK;
		}
	}

	// force the string to end in a null character if we happen to actually fill the buffer
	pResult[maxResult-1] = '\0';

	return S_OK;
}

bool ReadMemoryHelper(DEBUGHELPER * pHelper
						 , unsigned __int64 dlAddrToRead
						 , DWORD dwSizeToRead
						 , VOID* pBuffer
						 , unsigned __int64 dwAlignment
						 , DWORD * pOutSizeRead)
{
	HRESULT hr = S_OK;
	// the whole purpose of this function is to read memory across memory page boundaries
	if ((dlAddrToRead & ~(dwAlignment-1)) == (dlAddrToRead+dwSizeToRead & ~(dwAlignment-1)))
	{
		// common case - all in one read
		hr = pHelper->ReadDebuggeeMemoryEx(pHelper, dlAddrToRead, dwSizeToRead, pBuffer, pOutSizeRead);

		return hr == S_OK;
	}

	DWORD nGot;
	unsigned __int64 dlPageMask = ~(dwAlignment-1);

	*pOutSizeRead = 0;

	// we bridge a page boundary, so we have to split our read up into parts
	while ((dwSizeToRead > 0) && (hr == S_OK))
	{
		unsigned __int64 dlStartAddr = dlAddrToRead;
		unsigned __int64 dlEndAddr = dlStartAddr+dwSizeToRead;

		// If the data to read bridges one of our page boundaries (as specified by the Alignment argument)
		//   Then clip the end of the read to the end of the page. We'll loop back around to get more later.
		if ((dlStartAddr & dlPageMask) != (dlEndAddr & dlPageMask))
		{ 
			dlEndAddr = (dlStartAddr + dwAlignment) & dlPageMask;
		}

		DWORD dwReadSize = static_cast<DWORD>(dlEndAddr - dlStartAddr);

		hr = pHelper->ReadDebuggeeMemoryEx(pHelper, dlAddrToRead, dwReadSize, pBuffer, &nGot);
		if (hr == S_OK)
		{
			dwSizeToRead -= dwReadSize;
			pBuffer = (VOID*)(reinterpret_cast<unsigned __int64>(pBuffer) + dwReadSize);
			*pOutSizeRead += nGot;
			dlAddrToRead += dwReadSize;
		}
	}

	return hr == S_OK;
}

// Convenience function to convert 4- or 8-byte values in either endian into value
//  8-byte little endian values. 
unsigned __int64 ConvertTargetValueToLocalValue(unsigned __int64 Value
												, unsigned int uiSizeOfValue
												, bool bIsBigEndianTarget)
{
	if (uiSizeOfValue == 8)
	{
		if (bIsBigEndianTarget) 
		{
			Value = FlipEndian64(Value);
		}
	}
	else if (uiSizeOfValue == 4)
	{
		if (bIsBigEndianTarget) 
		{
			Value = (unsigned __int64)FlipEndian32((DWORD)Value);
		}

		// truncate the Most significant bits of pFNamesData so we have a valid pointer
		Value = Truncate64BitValueTo32BitValue(Value);
	}

	return Value;
}

unsigned __int64 Truncate64BitValueTo32BitValue(unsigned __int64 Value)
{
	return Value & 0x00000000FFFFFFFF;
}


unsigned int FlipEndian32(unsigned int Data)
{
	unsigned int Temp = Data;

	Data =  ((Temp>>24) & 0x000000FF) |
			((Temp>>8)  & 0x0000FF00) |
			((Temp<<8)  & 0x00FF0000) |
			((Temp<<24) & 0xFF000000) ;

	return Data;
}

unsigned __int64 FlipEndian64(unsigned __int64 Data)
{
	unsigned __int64 Temp = Data;

	Data =  ((Temp>>56) & 0x00000000000000FF) |
			((Temp>>40) & 0x000000000000FF00) |
			((Temp>>24) & 0x0000000000FF0000) |
			((Temp>>8)  & 0x00000000FF000000) | 
			((Temp<<8)  & 0x000000FF00000000) | 
			((Temp<<24) & 0x0000FF0000000000) | 
			((Temp<<40) & 0x00FF000000000000) | 
			((Temp<<56) & 0xFF00000000000000) ;

	return Data;
}